from __future__ import print_function
"""
Variables
"""

"""
RCSid:
	$Id: vars.py,v 1.26 2022/05/29 07:44:57 sjg Exp $

	@(#) Copyright (c) 2014-2022 Simon J. Gerraty

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import fnmatch
import os
import os.path
import re
import sys
import subprocess

def unescape(s,x=''):
    """handle common escape chars"""
    ra = []
    i = 0
    n = len(s)
    while i < n:
        c = s[i]
        i += 1
        if c == '\\':
            if s[i] == '\\' or s[i] in x:
                ra.append(s[i])
            elif s[i] == 'n':
                ra.append('\n')
            elif s[i] == 't':
                ra.append('\t')
            i += 1
            continue
        ra.append(c)
    return ''.join(ra)

def str_subst(s,p,r,gf=False):
    """substitute r for p in s

    if gf is True replace every instance, otherwise just the first
    """
    ra = []
    if p.startswith('^'):
        if s.startswith(p[1:]):
           ra.append(r)
           ra.append(s[len(p) - 1:])
        else:
            return s
    elif p.endswith('$'):
        if s.endswith(p[0:-1]):
            ra.append(s[0:-len(p)])
            ra.append(r)
        else:
            return s
    else:
        if gf:
            n = len(s)
        else:
            n = 1
        while n > 0:
            x = s.find(p)
            if x < 0:
                return s
            n -= 1
            if x > 0:
                ra.append(s[0:x])
            ra.append(r)
            ra.append(s[x+len(p):])
            s = ''.join(ra)
            ra = []
        return s
    return ''.join(ra)

def re_subst(s,p,r,gf=False):
    """substitute r for p in s

    if gf is True replace every instance, otherwise just the first
    """
    mp = re.compile(unescape(p,'()'))
    
    if gf:
        n = 0
    else:
        n = 1
    return mp.sub(r, s, n)

Braces = {
    '{': '}',
    '(': ')',
    '<': '>',
    '[': ']'
    }

def find_ec(s,t=None):
    """find the index of end char that matches the first or t if set"""
    n=0
    sc = s[0]
    if sc in Braces:
        ec = Braces[sc]
    else:
        ec = None
    for i in range(len(s)):
        c = s[i]
        if c == sc:
            n += 1
        elif c == ec:
            n -= 1
            if n == 0:
                return i
        elif c == t and n == 1:
            return i
    return -1

class varMod(object):
    """track state of var during a series of mods"""
    def __init__(self, name, val):
        self.name = name
        if val:
            self.va = val.split()
        else:
            self.va = []
        self.val =  None
        self.sep = ' '
        
    def get_str(self):
        if not self.val is None:
            return self.val
        return self.sep.join(self.va)

    def get_words(self):
        if not self.val is None:
            self.va = self.val.split()
            self.val = None
        return self.va

    def set_words(self, va=None):
        if not va is None:
            self.va = va
        elif self.val:
            self.va = self.val.split()
        self.val = None

    def set_str(self, str=None):
        if not str is None:
            self.val = str
        else:
            self.val = self.sep.join(self.va)

    def set_sep(self, sep):
        self.sep = sep


class Vars(object):
    """Class for variable storage and substiutions

    The syntax is similar to bmake(1).
    """
    
    GET_NO_ENV=1
    
    def __init__(self, vars={}):
        self.vars = vars;
        self.set('.newline', '\n');
        self.set('.', vars.get('progname', sys.argv[0]))
        self.set('progname', '${.:T}', ':=')
        self.set('progdir',  '${.:H:tA}', ':=')
        self.delete('.')

    def set(self, var, val, op='=', flags=0):
        """set var to val according to op

        ?= we ignore val if var already set
        += we append val to existing value
        =  we override any current value
        := as for '=' but expand val first
        != run val and use output as value
        """
        if var.find('$') >= 0:
            var = self.subst(var)
        if op == '=':
            self.vars[var] = val
        elif op == ':=':
            self.vars[var] = self.subst(val)
        elif op == '?=':
            if self.get(var, flags=flags):
                return
            self.vars[var] = val
        elif op == '+=':
            self.append(var,val)
        elif op == '!=':
            # PITA to get this to work with python 2 and 3
            r = subprocess.Popen(val.split(), universal_newlines=True,
                                 stdout=subprocess.PIPE)
            self.vars[var] = ''.join(r.stdout.readlines()).strip()

    def append(self, var, val):
        if not var in self.vars:
            self.vars[var] = val
        else:
            self.vars[var] += ' ' + val

    def delete(self, var):
        """remove var"""
        if var.find('$') >= 0:
            var = self.subst(var)
        if var in self.vars:
            del self.vars[var]

    def get(self, var, d=None, flags=0):
        if var.find('$') >= 0:
            var = self.subst(var)
        if var in self.vars:
            return self.vars[var]
        if (flags & self.GET_NO_ENV) == 0:
            return os.getenv(var, d)
        return d

    def dump(self, file=sys.stdout):
        for k,v in self.vars.items():
            print('{}={}'.format(k,v), file=file)
            
    def modify(self, var, val, mods, debug=0):
        """Apply a series of modidiers"""

        if debug:
            print('apply: {} to: {}'.format(mods,val))
        vm = varMod(var,val)
        while True:
            if len(mods) == 0:
                break
            # corner cases first - might contain ':'
            if mods.startswith('$') and mods[1] in '{(':
                ei = find_ec(mods[1:])
                if ei < 0:
                    raise ValueError(mods)
                x = self.subst(mods[:ei+2])
                mods = x + mods[ei+2:]
            if mods.startswith('S') or mods.startswith('C'):
                S,old,new,flags = mods.split(mods[1],3)
                if flags.find(':') >= 0:
                    flags,mods = flags.split(':', 1)
                else:
                    mods = ''
                wf = one = gf = False
                nv = []
                if 'W' in flags:
                    wf = True
                if '1' in flags:
                    one = True
                if 'g' in flags:
                    gf = True
                if wf:
                    wl = [vm.get_str()]
                else:
                    wl = vm.get_words()
                done = False
                for w in wl:
                    if not done:
                        if S == 'C':
                            nv.append(re_subst(w, old, new, gf))
                        else:
                            nv.append(str_subst(w, old, new, gf))
                        if one:
                            done = True
                    else:
                        nv.append(w)
                # corner case!
                if old == ':' and new == ' ':
                    nnv = []
                    for w in nv:
                        nnv += w.split()
                    nv = nnv
                vm.set_words(nv)
                continue
            if mods.startswith('ts:'):
                vm.set_sep(':')
                mods = mods[3:]
                continue
            if mods.startswith('@'):
                j,tv,ts,r = mods.split('@', 3)
                # if ts is sane we are good
                i = ni = 0
                while True:
                    ni = ts[i:].find('${')
                    if ni < 0:
                        ni = ts[i:].find('$(')
                    if ni < 0:
                        break
                    i = ni + 1
                    ei = find_ec(ts[i:])
                    if ei < 0:
                        break
                    i = ei + 1
                if i > 0 and ei < 0:
                    # parsing is too complicated! ;-)
                    raise ValueError(mods)
                if r.startswith(':'):
                    mods = r[1:]
                else:
                    mods = ''
                # temp vars context
                tvars = Vars(self.vars)
                nv = []
                for w in vm.get_words():
                    tvars.set(tv, w)
                    nv.append(tvars.subst(ts, debug))
                vm.set_words(nv)
                continue
            if mods.find(':') >= 0:
                m,mods = mods.split(':', 1)
            else:
                m = mods
                mods = ''
            
            if m.startswith('['):
                # the only non-numeric we support is '#'
                x = m[1:-1].split('..')
                w = vm.get_words()
                if len(x) == 1:
                    if x[0] == '#':
                        vm.set_str('{}'.format(len(w)))
                        continue
                    if x[0] == '*' or x[0] == '0':
                        vm.set_str()
                        continue
                    if x[0] == '@':
                        vm.set_words()
                        continue
                    i = int(x[0])
                    if i < 0:
                        vm.set_str(w[i])
                    elif i > 0:
                        vm.set_str(w[i-1])
                    continue
                # a range - a bit messy since we are matching make
                ix = map(int, x)
                if ix[0] > 0:
                    ix[0] -= 1
                if ix[1] == -1:
                    vm.set_words(w[ix[0]:])
                elif ix[0] == -1 and ix[1] == 1:
                    vm.set_words(w[-1::-1])
                else:
                    vm.set_words(w[ix[0]:ix[1]])
            elif m.startswith('D'):
                if self.get(var):
                    vm.set_str(m[1:])
            elif m == 'E':
                nv = []
                for w in vm.get_words():
                    ext = os.path.splitext(w)[1]
                    if ext:
                        nv.append(ext[1:])
                    else:
                        nv.append('')
                vm.set_words(nv)
            elif m == 'H':
                nv = map(lambda s: os.path.dirname(s), vm.get_words())
                vm.set_words(nv)
            elif m == 'L':
                vm.set_str(var)
            elif m.startswith('M'):
                nv = []
                p = m[1:]
                for w in vm.get_words():
                    if fnmatch.fnmatch(w, p):
                        nv.append(w)
                vm.set_words(nv)
            elif m.startswith('N'):
                nv = []
                p = m[1:]
                for w in vm.get_words():
                    if not fnmatch.fnmatch(w, p):
                        nv.append(w)
                vm.set_words(nv)
            elif m == 'O':
                nv = vm.get_words()
                vm.set_words(sorted(nv))
            elif m == 'R':
                nv = map(lambda s: os.path.splitext(s)[0], vm.get_words())
                vm.set_words(nv)
            elif m == 'T':
                nv = map(lambda s: os.path.basename(s), vm.get_words())
                vm.set_words(nv)
            elif m == 'tA':
                nv = map(lambda s: os.path.realpath(s), vm.get_words())
                vm.set_words(nv)
            elif m == 'tl':
                vm.set_str(vm.get_str().lower())
            elif m.startswith('ts'):
                vm.set_sep(unescape(m[2:]))
            elif m == 'tu':
                vm.set_str(vm.get_str().upper())
            elif m.startswith('U'):
                if not self.get(var):
                    vm.set_str(m[1:])
            elif m == 'u':
                nv = []
                lw = ''
                for w in vm.get_words():
                    if w != lw:
                        nv.append(w)
                        lw = w
                vm.set_words(nv)
        return vm.get_str()

    def subst(self, s, debug=0):
        """substitute any var references"""
        
        if debug:
            print('subst: {}'.format(s))
        ra = []
        i = 0
        n = len(s)
        ec = None
        while i < n:
            if i == 0:
                rs = s
            else:
                rs = s[i:]
            x = rs.find('$')
            if x < 0:
                if i == 0:
                    return s
                ra.append(rs)
                break
            if x > 0:
                ra.append(rs[:x])
                i += x
            c = s[i]
            i += 1
            if c == '$':
                if s[i] == '$':
                    ra.append(c)
                    i += 1
                    continue
                if s[i] not in '{(':
                    # simple single char
                    v = s[i]
                    i += 1
                    ra.append(self.get(v,''))
                    continue
                rs = s[i:]
                m = find_ec(rs, ':')
                if rs[m] == ':':
                    ei = find_ec(rs)
                else:
                    ei = m
                si = rs.find('$')
                if si > 0 and si < m:
                    # variable within the variable name we need to recurse
                    v = self.subst(rs[1:m])
                else:
                    v = rs[1:m]
                val = self.get(v,'')
                if val.find('$') >= 0:
                    val = self.subst(val)
                if m > 0 and m < ei:
                    val = self.modify(v, val, rs[m+1:ei], debug)
                ra.append(val)
                i += ei + 1
            else:
                ra.append(c)
        r = ''.join(ra)
        if debug:
            print('subst-> {}'.format(r))
        if r.find('$') >= 0:
            r = self.subst(r, debug)
        return r

                
if __name__ == '__main__':
    import getopt
    
    vars = Vars()

    opts,args = getopt.getopt(sys.argv[1:], 'a:dr:s:')
    debug = 0
    for o,a in opts:
        if o == '-d':
            debug += 1
        elif o == '-a':
            k,v = a.split('=', 1)
            vars.append(k,v)
        elif o == '-r':
            for line in open(a).readlines():
                if line.startswith('#'):
                    continue
                k,v = line.strip().split('=', 1)
                vars.set(k.strip(), v.strip())
        elif o == '-s':
            k,v = a.split('=', 1)
            vars.set(k,v)

    if debug:
        vars.dump()

    for a in args:
        print('{}->{}'.format(a, vars.subst(a, debug)))
        


